"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SwapToLP = void 0;
const web3_js_1 = require("@solana/web3.js");
const core_1 = require("@jup-ag/core");
const cross_fetch_1 = __importDefault(require("cross-fetch"));
const raydium_sdk_1 = require("@raydium-io/raydium-sdk");
const spl_token_1 = require("@solana/spl-token");
/**
 * SwapToLP class for swapping SPL tokens to LP tokens through rebalancing and staking.
 */
class SwapToLP {
    /**
     * Constructor for the SwapToLP class. Builds and returns a SwapToLP object.
     */
    constructor(connection, jupiter) {
        this.connection = connection;
        this.jupiter = jupiter;
    }
    /**
     * Initializes a SwapToLP object.
     */
    static build(connection, user) {
        return __awaiter(this, void 0, void 0, function* () {
            // 1. Fetch fee token accounts
            const feeAccounts = yield (0, core_1.getPlatformFeeAccounts)(connection, new web3_js_1.PublicKey('6nYhSkKDU1X8ZuygrrYYdkU6jWT8mWWHbN2D3TvVrqxp'));
            const platformFeeAndAccounts = {
                feeBps: 50,
                feeAccounts: feeAccounts
            };
            // 2. Instantiate jupiter instance
            const jupiter = yield core_1.Jupiter.load({
                connection,
                cluster: 'mainnet-beta',
                user: user,
                platformFeeAndAccounts
            });
            // 3. Instantiate and return SwapToLP object
            return new SwapToLP(connection, jupiter);
        });
    }
    /**
     * Generates a set of unsigned transactions for swapping a single SPL token to base/quote tokens required for LP. The transactions include setup transactions for the output tokens based on the LP token's base and quote mints.
     * @param inputTokenAmount
     * @param inputTokenMint
     * @param lpTokenMint
     */
    swap(inputTokenAmount, inputTokenMint, lpTokenMint) {
        return __awaiter(this, void 0, void 0, function* () {
            var transactions = [];
            // 1. Fetch pool and token info
            const poolInfo = yield this.getPoolInfo(lpTokenMint);
            const baseMint = poolInfo.baseMint;
            const quoteMint = poolInfo.quoteMint;
            const baseInfo = yield this.getTokenInfo(baseMint);
            const quoteInfo = yield this.getTokenInfo(quoteMint);
            const inputTokenInfo = yield this.getTokenInfo(inputTokenMint.toBase58());
            const inputTokenPrice = yield this.getTokenPrice(inputTokenMint.toBase58(), inputTokenInfo.decimals);
            const basePrice = yield this.getTokenPrice(poolInfo.baseMint, baseInfo.decimals);
            const quotePrice = yield this.getTokenPrice(poolInfo.quoteMint, quoteInfo.decimals);
            var routes = [];
            // 2. Convert 1/2 inputTokenAmount in USD to token A
            // 2. Skip if input token = base token
            if (inputTokenMint.toBase58() !== poolInfo.baseMint) {
                // 2.1. Find routes for swapping input to base token
                const baseRoutes = yield this.jupiter.computeRoutes({
                    inputMint: inputTokenMint,
                    outputMint: new web3_js_1.PublicKey(baseMint),
                    inputAmount: inputTokenAmount / 2,
                    slippage: 1,
                    forceFetch: true
                });
                routes.push(baseRoutes.routesInfos[0]);
                // 2.2. Generate transactions
                const baseExchange = yield this.jupiter.exchange({ routeInfo: baseRoutes.routesInfos[0] });
                // 2.3. Add transactions
                if (baseExchange.transactions.setupTransaction) {
                    transactions.push(baseExchange.transactions.setupTransaction);
                }
                if (baseExchange.transactions.swapTransaction) {
                    transactions.push(baseExchange.transactions.swapTransaction);
                }
                if (baseExchange.transactions.cleanupTransaction) {
                    transactions.push(baseExchange.transactions.cleanupTransaction);
                }
            }
            // 3. Convert 1/2 inputTokenAmount in USD to token B
            // 3. Skip if input token = quote token
            if (inputTokenMint.toBase58() !== poolInfo.quoteMint) {
                // 3.1 Find routes for swapping input to quote token
                const quoteRoutes = yield this.jupiter.computeRoutes({
                    inputMint: inputTokenMint,
                    outputMint: new web3_js_1.PublicKey(quoteMint),
                    inputAmount: inputTokenAmount / 2,
                    slippage: 1,
                    forceFetch: true
                });
                routes.push(quoteRoutes.routesInfos[0]);
                // 3.2. Generate transactions
                const quoteExchange = yield this.jupiter.exchange({ routeInfo: quoteRoutes.routesInfos[0] });
                // 3.3. Add transactions
                if (quoteExchange.transactions.setupTransaction) {
                    transactions.push(quoteExchange.transactions.setupTransaction);
                }
                if (quoteExchange.transactions.swapTransaction) {
                    transactions.push(quoteExchange.transactions.swapTransaction);
                }
                if (quoteExchange.transactions.cleanupTransaction) {
                    transactions.push(quoteExchange.transactions.cleanupTransaction);
                }
            }
            return {
                transactions,
                poolInfo,
                prices: {
                    inputTokenPrice,
                    basePrice,
                    quotePrice
                },
                tokenInfo: {
                    inputTokenInfo,
                    baseInfo,
                    quoteInfo
                },
                routes
            };
        });
    }
    /**
     * Generate an unsigned transactions with instruction to add liquidity to LP.
     * @param lpTokenMint
     * @param poolInfo
     * @param tokenAAmount
     */
    addLiquidity(lpTokenMint, poolInfo, tokenAAmount, user) {
        return __awaiter(this, void 0, void 0, function* () {
            // 1. Create new transaction
            var addLiquidityTx = new web3_js_1.Transaction();
            // 2. Find user's token account based on LP mint
            console.log('Fetching LP mint addresss');
            const usersLpTokenAddress = yield raydium_sdk_1.Spl.getAssociatedTokenAccount({ mint: lpTokenMint, owner: user });
            // 3. Fetch account info for user's 
            const associatedAccount = yield this.connection.getAccountInfo(usersLpTokenAddress);
            if (associatedAccount === null) {
                var ix = yield spl_token_1.Token.createAssociatedTokenAccountInstruction(spl_token_1.ASSOCIATED_TOKEN_PROGRAM_ID, spl_token_1.TOKEN_PROGRAM_ID, lpTokenMint, usersLpTokenAddress, user, user);
                addLiquidityTx.add(ix);
            }
            console.log('Converting info to keys');
            const poolKeys = (0, raydium_sdk_1.jsonInfo2PoolKeys)(poolInfo);
            // get pool quote/base token decimals
            console.log('Fetching pool info');
            const raydiumPoolInfo = yield raydium_sdk_1.Liquidity.fetchInfo({ connection: this.connection, poolKeys });
            console.log('Cerating token 1');
            const baseCurrency = new raydium_sdk_1.Token(poolInfo.baseMint, raydiumPoolInfo.baseDecimals);
            console.log('Cerating token 2');
            const quoteCurrency = new raydium_sdk_1.Token(poolInfo.quoteMint, raydiumPoolInfo.quoteDecimals);
            console.log('Cerating token amount');
            const tokenAmountTokenA = new raydium_sdk_1.TokenAmount(baseCurrency, Math.round(tokenAAmount), true);
            console.log('Computing amount');
            const { maxAnotherAmount } = raydium_sdk_1.Liquidity.computeAnotherAmount({
                poolKeys,
                poolInfo: raydiumPoolInfo,
                amount: tokenAmountTokenA,
                anotherCurrency: quoteCurrency,
                slippage: new raydium_sdk_1.Percent(0),
            });
            console.log('Creating ix');
            const tx = yield raydium_sdk_1.Liquidity.makeAddLiquidityInstructionV4({
                poolKeys: poolKeys,
                userKeys: {
                    baseTokenAccount: yield raydium_sdk_1.Spl.getAssociatedTokenAccount({ mint: new web3_js_1.PublicKey(poolInfo.baseMint), owner: user }),
                    quoteTokenAccount: yield raydium_sdk_1.Spl.getAssociatedTokenAccount({ mint: new web3_js_1.PublicKey(poolInfo.quoteMint), owner: user }),
                    lpTokenAccount: usersLpTokenAddress,
                    owner: user
                },
                baseAmountIn: tokenAmountTokenA.raw,
                quoteAmountIn: maxAnotherAmount.raw,
                fixedSide: "base"
            });
            addLiquidityTx.add(tx);
            console.log('Creating final tx');
            addLiquidityTx.recentBlockhash = (yield this.connection.getRecentBlockhash()).blockhash;
            addLiquidityTx.feePayer = user;
            console.log('Created tx');
            return addLiquidityTx;
        });
    }
    /**
     * Helper method to bundling `swapToBaseQuote` and `addLiquidity` together.
     * @param inputTokenAmount
     * @param inputTokenMint
     * @param lpTokenMint
     * @returns Array of transactions
     */
    swapAndAddLiquidity(inputTokenAmount, inputTokenMint, lpTokenMint, user) {
        return __awaiter(this, void 0, void 0, function* () {
            // 1. Swap input token to base and quote tokens for LP
            const { transactions, poolInfo, prices, tokenInfo, routes } = yield this.swap(inputTokenAmount, inputTokenMint, lpTokenMint);
            // 2. Calculate quote amount to stake
            const inputTokenUSD = (inputTokenAmount / (Math.pow(10, tokenInfo.inputTokenInfo.decimals)) * prices.inputTokenPrice);
            const quoteAmount = (inputTokenUSD / 2) / prices.basePrice * Math.pow(10, tokenInfo.baseInfo.decimals);
            // 3. Stake base and quote tokens
            const addLiquidityTx = yield this.addLiquidity(lpTokenMint, poolInfo, quoteAmount, user);
            return {
                transactions: [...transactions, addLiquidityTx],
                routes: routes
            };
        });
    }
    /**
     * Get route info.
     * @param inputTokenAmount
     * @param inputTokenMint
     * @param lpTokenMint
     */
    getRouteInfo(inputTokenAmount = 2, inputTokenMint, lpTokenMint) {
        return __awaiter(this, void 0, void 0, function* () {
            var routes = [];
            const poolInfo = yield this.getPoolInfo(lpTokenMint);
            const baseMint = poolInfo.baseMint;
            const quoteMint = poolInfo.quoteMint;
            if (inputTokenMint.toBase58() !== poolInfo.baseMint) {
                const baseRoutes = yield this.jupiter.computeRoutes({
                    inputMint: inputTokenMint,
                    outputMint: new web3_js_1.PublicKey(baseMint),
                    inputAmount: inputTokenAmount / 2,
                    slippage: 1,
                    forceFetch: true
                });
                routes.push(baseRoutes.routesInfos[0]);
            }
            if (inputTokenMint.toBase58() !== poolInfo.quoteMint) {
                const quoteRoutes = yield this.jupiter.computeRoutes({
                    inputMint: inputTokenMint,
                    outputMint: new web3_js_1.PublicKey(quoteMint),
                    inputAmount: inputTokenAmount / 2,
                    slippage: 1,
                    forceFetch: true
                });
                routes.push(quoteRoutes.routesInfos[0]);
            }
            return routes;
        });
    }
    /**
     * Get pool info for LP token mint.
     * @param lpTokenMint - LP token mint
     */
    getPoolInfo(lpTokenMint) {
        return __awaiter(this, void 0, void 0, function* () {
            const raydiumURL = 'https://sdk.raydium.io/liquidity/mainnet.json';
            const response = yield (0, cross_fetch_1.default)(raydiumURL);
            const pools = yield response.json();
            const allPools = pools.official.concat(pools.unOfficial);
            const pool = allPools.find(p => p.lpMint === lpTokenMint.toBase58());
            if (!pool) {
                throw new Error('Pool not found');
            }
            return pool;
        });
    }
    /**
     * Get token info for a given mint.
     * @param tokenMint - token mint
     */
    getTokenInfo(tokenMint) {
        return __awaiter(this, void 0, void 0, function* () {
            const raydiumURL = 'https://raw.githubusercontent.com/solana-labs/token-list/main/src/tokens/solana.tokenlist.json';
            const response = yield (0, cross_fetch_1.default)(raydiumURL);
            const tokens = yield response.json();
            const tokenInfo = tokens.tokens.find(p => p.address === tokenMint);
            return tokenInfo;
        });
    }
    /**
     * Fetch user balances for 3 given mints.
     * @param tokenAMint
     * @param tokenBMint
     * @param lpTokenMint
     */
    getUserBalances(tokenAMint, tokenBMint, lpTokenMint, user) {
        return __awaiter(this, void 0, void 0, function* () {
            const tokenAAddress = yield raydium_sdk_1.Spl.getAssociatedTokenAccount({ mint: tokenAMint, owner: user });
            const tokenABalance = (yield this.connection.getTokenAccountBalance(tokenAAddress)).value.amount;
            const tokenBAddress = yield raydium_sdk_1.Spl.getAssociatedTokenAccount({ mint: tokenBMint, owner: user });
            const tokenBBalance = (yield this.connection.getTokenAccountBalance(tokenBAddress)).value.amount;
            const lpTokenAddress = yield raydium_sdk_1.Spl.getAssociatedTokenAccount({ mint: lpTokenMint, owner: user });
            const lpTokenBalance = (yield this.connection.getTokenAccountBalance(lpTokenAddress)).value.amount;
            return {
                tokenAAddress: tokenAAddress,
                tokenABalance: Number(tokenABalance),
                tokenBAddress: tokenBAddress,
                tokenBBalance: Number(tokenBBalance),
                lpTokenAddress: lpTokenAddress,
                lpTokenBalance: Number(lpTokenBalance)
            };
        });
    }
    /**
     * Fetch user balances for an LP mint and its base and quote tokens.
     * @param lpTokenMint
     */
    getUserBalancesFromLp(lpTokenMint, user) {
        return __awaiter(this, void 0, void 0, function* () {
            const poolInfo = yield this.getPoolInfo(lpTokenMint);
            const tokenAMint = poolInfo.baseMint;
            const tokenBMint = poolInfo.quoteMint;
            return this.getUserBalances(new web3_js_1.PublicKey(tokenAMint), new web3_js_1.PublicKey(tokenBMint), lpTokenMint, user);
        });
    }
    /**
     * Get token price for a given mint.
     * @param tokenMint
     * @param tokenDecimals
     */
    getTokenPrice(tokenMint, tokenDecimals) {
        return __awaiter(this, void 0, void 0, function* () {
            const jupiterURL = `https://quote-api.jup.ag/v1/quote?inputMint=${tokenMint}&outputMint=EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v&amount=${Math.pow(10, tokenDecimals)}`;
            const response = yield (0, cross_fetch_1.default)(jupiterURL);
            const jsonReponse = yield response.json();
            const outAmount = jsonReponse.data[0].outAmount / 1e6;
            return outAmount;
        });
    }
}
exports.SwapToLP = SwapToLP;
